(function() {
  /*
    Node index - Main Library file
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var DEFAULT_OPTIONS, Index, step, utils;
  step = require('step');
  utils = require('./index/utils');
  DEFAULT_OPTIONS = {
    sort: function(a, b) {
      if ((a === null) || (a < b)) {
        return -1;
      } else {
        if (a === b) {
          return 0;
        } else {
          return 1;
        }
      }
    },
    order: 64
  };
  /*
    Class @constructor
  */
  Index = exports.Index = function(options) {
    options = utils.merge(DEFAULT_OPTIONS, options);
    this.storage = options.storage || require('./index/memory-storage').createStorage();
    this.order = options.order, this.sort = options.sort, this.conflictManager = options.conflictManager;
    this.order--;
    this.lockQueue = [];
    return this;
  };
  /*
    Wrapper for class @constructor
  */
  exports.createIndex = function(options) {
    return new Index(options);
  };
  /*
   Get functionality
  */
  Index.prototype.get = require('./index/core/get').get;
  Index.prototype.traverse = require('./index/core/get').traverse;
  Index.prototype.rangeGet = require('./index/core/get').rangeGet;
  /*
    Set functionality
  */
  Index.prototype.set = require('./index/core/set').set;
  /*
    Unset functionality
  */
  Index.prototype.unset = require('./index/core/unset').unset;
  /*
    Compaction functionality
  */
  Index.prototype.compact = require('./index/core/compact').compact;
  /*
    Bulk functionality
  */
  Index.prototype.bulk = require('./index/core/bulk').bulk;
  /*
    Lock functionality
  */
  Index.prototype.lock = function(fn) {
    if (this.locked) {
      this.lockQueue.push(fn);
      return true;
    }
    this.locked = true;
    return false;
  };
  /*
    Release lock functionality
  */
  Index.prototype.releaseLock = function() {
    var fn;
    if (!this.locked) {
      return;
    }
    this.locked = false;
    fn = this.lockQueue.shift();
    if (!fn && this.lockQueue.length <= 0) {
      return;
    }
    return process.nextTick(fn);
  };
  /*
    Export storages
  */
  exports.storage = {
    memory: require('./index/memory-storage'),
    file: require('./index/file-storage')
  };
}).call(this);
