(function() {
  /*
    Bulk set/unset functionality for Node index module
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var splitPage, step, utils;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  step = require('step');
  utils = require('../../index/utils');
  /*
    Bulk set/unset
  
    kvs = [
      [key, value, 1] - set
      [key] - unset
    ]
  */
  exports.bulk = function(kvs, _callback) {
    var callback, conflictManager, conflicts, iterate, order, recSplit, sort, storage, that;
    that = this;
    sort = this.sort;
    order = this.order;
    storage = this.storage;
    conflictManager = this.conflictManager;
    if (this.lock(__bind(function() {
      return this.bulk(kvs, _callback);
    }, this))) {
      return;
    }
    callback = __bind(function(err, data) {
      this.releaseLock();
      return process.nextTick(function() {
        return _callback && _callback(err, conflicts);
      });
    }, this);
    conflicts = [];
    kvs = [].concat(kvs);
    kvs = kvs.sort(function(a, b) {
      return sort(a[0], b[0]);
    });
    if (!(kvs.length > 0)) {
      return callback(null, []);
    }
    iterate = function(callback, kvs) {
      var _iterate;
      return _iterate = function(err, page) {
        var index, item, kv, kv_index, _kvs;
        if (err) {
          return callback(err);
        }
        kv = kvs.shift();
        if (!kv) {
          splitPage(page, order, storage, callback);
          return;
        }
        index = utils.search(page, sort, kv[0]);
        item = page[index];
        if (item && !item[2]) {
          kvs.unshift(kv);
          if (page[index + 1]) {
            kv_index = utils.search(kvs, sort, page[index + 1][0]);
            if (!kvs[kv_index]) {
              kv_index++;
            } else if (sort(kvs[kv_index][0], page[index + 1][0]) !== 0) {
              kv_index++;
            }
            _kvs = kvs.splice(0, kv_index);
          } else {
            _kvs = kvs;
          }
          return step(function() {
            return storage.read(item[1], this.parallel());
          }, iterate(function(err, kps) {
            if (err) {
              return callback(err);
            }
            if (kps.length === 0) {
              page.splice(index, 1);
            } else {
              kps[0][0] = item[0];
              page.splice.apply(page, [index, 1].concat(kps));
            }
            return _iterate(null, page);
          }, _kvs));
        } else {
          if (!item || (sort(item[0], kv[0])) < 0) {
            if (kv[2]) {
              index = index === null ? 0 : index + 1;
              page.splice(index, 0, kv);
              storage.write([kv[1]], function(err, pos) {
                kv[1] = pos;
                return process.nextTick(function() {
                  return _iterate(null, page);
                });
              });
              return;
            } else {

            }
          } else {
            if (!kv[2]) {
              page.splice(index, 1);
            } else {
              if (conflictManager) {
                conflicts.push(kv[0]);
              } else {
                conflicts.push(kv[0]);
              }
            }
          }
          return process.nextTick(function() {
            return _iterate(null, page);
          });
        }
      };
    };
    return step(function() {
      return storage.readRoot(this.parallel());
    }, iterate(recSplit = function(err, page) {
      if (err) {
        return callback(err);
      }
      if (page.length === 0) {
        return storage.write([], function(err, pos) {
          if (err) {
            return callback(err);
          }
          return storage.writeRoot(pos, callback);
        });
      } else if (page.length === 1) {
        return storage.writeRoot(page[0][1], callback);
      } else {
        return splitPage(page, order, storage, recSplit);
      }
    }, kvs));
  };
  splitPage = function(page, order, storage, callback) {
    var i, len, pages;
    if (page.length === 0) {
      pages = [];
    } else if (page.length <= order) {
      pages = [page];
    } else {
      pages = [];
      i = 0;
      len = page.length;
      while (len > order) {
        len = len >> 1;
      }
      while (page.length > 0) {
        pages.push(page.splice(0, len));
      }
    }
    return step(function() {
      var group, page, pre_page, _i, _len;
      group = this.group();
      for (_i = 0, _len = pages.length; _i < _len; _i++) {
        page = pages[_i];
        pre_page = [].concat(page);
        if (!page[0][2]) {
          pre_page[0] = [].concat(pre_page[0]);
          pre_page[0][0] = null;
        }
        storage.write(pre_page, group());
      }
    }, function(err, pages_pos) {
      if (err) {
        return callback(err);
      }
      pages = pages_pos.map(function(pos, i) {
        if (i === 0) {
          return [null, pos];
        } else {
          return [pages[i][0][0], pos];
        }
      });
      return this.parallel()(null, pages);
    }, callback);
  };
}).call(this);
