(function() {
  /*
    Compaction for Node Index
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var step, utils;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  step = require('step');
  utils = require('../../index/utils');
  exports.compact = function(_callback) {
    var callback, iterate, storage;
    storage = this.storage;
    callback = __bind(function(err, data) {
      this.releaseLock();
      return process.nextTick(function() {
        return _callback && _callback(err, data);
      });
    }, this);
    if (this.lock(__bind(function() {
      return this.compact(_callback);
    }, this))) {
      return;
    }
    iterate = function(callback) {
      return function(err, page) {
        var fns, in_leaf;
        if (err) {
          return callback(err);
        }
        in_leaf = page[0] && page[0][2];
        fns = page.map(function(item) {
          return function() {
            return step(function() {
              storage.read(item[1], this.parallel(), in_leaf);
            }, function(err, data) {
              if (err) {
                throw err;
              }
              if (in_leaf) {
                storage.write(data, this.parallel(), in_leaf);
                return;
              }
              iterate(this.parallel())(null, data);
            }, function(err, new_pos) {
              if (err) {
                throw err;
              }
              item[1] = new_pos;
              this.parallel()(null);
            }, this.parallel());
          };
        });
        fns.push(function(err) {
          if (err) {
            throw err;
          }
          storage.write(page, this.parallel());
        });
        fns.push(callback);
        return step.apply(null, fns);
      };
    };
    return step(function() {
      if (storage.beforeCompact) {
        storage.beforeCompact(this.parallel());
      } else {
        this.parallel()(null);
      }
    }, function(err) {
      if (err) {
        throw err;
      }
      storage.readRoot(iterate(this.parallel()));
    }, function(err, new_root_pos) {
      if (err) {
        throw err;
      }
      storage.writeRoot(new_root_pos, this.parallel());
    }, function(err) {
      if (err) {
        throw err;
      }
      if (storage.afterCompact) {
        storage.afterCompact(this.parallel());
      } else {
        this.parallel()(null);
      }
    }, callback);
  };
}).call(this);
