(function() {
  /*
    Get functionality for Node Index module
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var step, utils;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  utils = require('../../index/utils');
  step = require('step');
  /*
    Get value by key
  */
  exports.get = function(key, callback) {
    var iterate, sort, storage;
    sort = this.sort;
    storage = this.storage;
    iterate = function(err, index) {
      var item, item_index, value;
      if (err) {
        return callback(err);
      }
      item_index = utils.search(index, sort, key);
      item = index[item_index];
      if (!item) {
        return callback('Not found');
      }
      value = item[1];
      if (item[2]) {
        if (sort(item[0], key) !== 0) {
          return callback('Not found');
        }
        return storage.read(value, function(err, value) {
          if (err) {
            return callback(err);
          }
          return callback(null, value[0]);
        });
      } else {
        return storage.read(value, iterate);
      }
    };
    return storage.readRoot(iterate);
  };
  /*
    Traverse btree
    filter can call callback w/ following result:
      true - if `traverse` should go deeper
      undefined - if `traverse` should skip element
      false - if `traverse` should stop traversing and return to higher level
  
    filter can
    @return promise.
  */
  exports.traverse = function(filter) {
    var promise, that;
    that = this;
    promise = new process.EventEmitter;
    filter = filter || function(kp, callback) {
      return callback(null, true);
    };
    process.nextTick(__bind(function() {
      var iterate, sort, storage;
      sort = this.sort;
      storage = this.storage;
      iterate = function(callback) {
        return function(err, page) {
          var asyncFilter, index, pagelen;
          if (err) {
            promise.emit('error', err);
            promise.emit('end');
            return;
          }
          index = -1;
          pagelen = page.length;
          asyncFilter = function() {
            var current;
            if (++index >= pagelen) {
              if (callback) {
                callback(null);
              } else {
                promise.emit('end');
              }
              return;
            }
            current = page[index];
            return filter.call(that, current, function(err, filter_value) {
              if (filter_value === true) {
                if (current[2]) {
                  return storage.read(current[1], function(err, value) {
                    promise.emit('data', value[0], current);
                    return asyncFilter();
                  });
                } else {
                  return storage.read(current[1], iterate(asyncFilter));
                }
              } else {
                if (filter_value === false) {
                  index = pagelen;
                }
                return asyncFilter();
              }
            });
          };
          return asyncFilter();
        };
      };
      return storage.readRoot(iterate());
    }, this));
    return promise;
  };
  /*
    Get in range
  */
  exports.rangeGet = function(start_key, end_key) {
    var promise, sort, traverse_promise;
    sort = this.sort;
    promise = new process.EventEmitter;
    traverse_promise = this.traverse(function(kp, callback) {
      var end_cmp, start_cmp;
      start_cmp = sort(kp[0], start_key);
      end_cmp = sort(kp[0], end_key);
      if (kp[2]) {
        if (start_cmp >= 0 && end_cmp <= 0) {
          return callback(null, true);
        }
        if (end_cmp > 0) {
          return callback(null, false);
        }
      } else {
        if (end_cmp <= 0) {
          return callback(null, true);
        }
        if (end_cmp > 0) {
          return callback(null, false);
        }
      }
      return callback(null);
    });
    traverse_promise.on('data', function(value) {
      return promise.emit('data', value);
    });
    traverse_promise.on('end', function() {
      return promise.emit('end');
    });
    return promise;
  };
}).call(this);
