(function() {
  /*
    Set functionality for Node index module
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var splitPage, step, utils;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  step = require('step');
  utils = require('../../index/utils');
  /*
    Set
  */
  exports.set = function(key, value, _callback) {
    var callback, conflictManager, iterate, order, sort, storage;
    sort = this.sort;
    order = this.order;
    storage = this.storage;
    conflictManager = this.conflictManager;
    if (this.lock(__bind(function() {
      return this.set(key, value, _callback);
    }, this))) {
      return;
    }
    callback = __bind(function(err, data) {
      this.releaseLock();
      return process.nextTick(function() {
        return _callback && _callback(err, data);
      });
    }, this);
    iterate = function(page, callback) {
      var item, item_index;
      item_index = utils.search(page, sort, key);
      item = page[item_index];
      if (item && !item[2]) {
        return step(function() {
          return storage.read(item[1], this.parallel());
        }, function(err, page) {
          if (err) {
            throw err;
          }
          return iterate(page, this.parallel());
        }, function(err, result) {
          if (err) {
            callback(err);
            return;
          }
          if (storage.isPosition(result)) {
            page[item_index][1] = result;
            return storage.write(page, callback);
          } else {
            page[item_index][1] = result.left_page;
            page.splice(item_index + 1, 0, [result.middle_key, result.right_page]);
            return splitPage(false, storage, order, page, callback);
          }
        });
      } else {
        return step(function() {
          if (item && sort(item[0], key) === 0) {
            if (!conflictManager) {
              throw 'Can\'t insert item w/ dublicate key';
            }
            step(function() {
              return storage.read(item[1], this.parallel());
            }, function(err, old_value) {
              if (err) {
                throw err;
              }
              this.parallel()(null, old_value);
              return conflictManager(old_value, value, this.parallel());
            }, this.parallel());
            return;
          }
          return this.parallel()(null, value);
        }, function(err, value, old_value) {
          if (err) {
            throw err;
          }
          item_index = item_index === null ? 0 : item_index + 1;
          return storage.write([value, old_value], this.parallel());
        }, function(err, value) {
          if (err) {
            callback(err);
            return;
          }
          page.splice(item_index, 0, [key, value, 1]);
          return splitPage(true, storage, order, page, callback);
        });
      }
    };
    return step(function() {
      return storage.readRoot(this.parallel());
    }, function(err, root) {
      if (err) {
        throw err;
      }
      return iterate(root, this.parallel());
    }, function(err, result) {
      if (err) {
        throw err;
      }
      if (storage.isPosition(result)) {
        return this.parallel()(null, result);
      } else {
        return storage.write([[null, result.left_page], [result.middle_key, result.right_page]], this.parallel());
      }
    }, function(err, new_root_pos) {
      if (err) {
        throw err;
      }
      return storage.writeRoot(new_root_pos, this.parallel());
    }, callback);
  };
  /*
    Check page length
    If exceed - split it into two and return left_page, right_page, middle_key
  */
  splitPage = function(in_leaf, storage, order, page, callback) {
    var mid_index, mid_key;
    if (page.length > order) {
      mid_index = page.length >> 1;
      mid_key = page[mid_index][0];
      return step(function() {
        var left_page, right_page;
        left_page = page.slice(0, mid_index);
        storage.write(left_page, this.parallel());
        right_page = page.slice(mid_index);
        if (!in_leaf) {
          right_page[0][0] = null;
        }
        return storage.write(right_page, this.parallel());
      }, function(err, left_page, right_page) {
        return callback(err, {
          left_page: left_page,
          middle_key: mid_key,
          right_page: right_page
        });
      });
    } else {
      return storage.write(page, callback);
    }
  };
}).call(this);
