(function() {
  /*
    Memory storage for Node Index Module
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var Position, Storage, isPosition, step, util;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  util = require('util');
  step = require('step');
  /*
    Class @constructor
  */
  Storage = exports.Storage = function(options) {
    this.data = [[]];
    this.root_pos = new Position(0);
    return this;
  };
  /*
    @constructor wrapper
  */
  exports.createStorage = function(options) {
    return new Storage(options);
  };
  Storage.prototype.isPosition = isPosition = function(pos) {
    return pos instanceof Position;
  };
  Storage.prototype.read = function(pos, callback) {
    if (!isPosition(pos)) {
      return callback('pos should be a valid position');
    }
    return process.nextTick(__bind(function() {
      return callback(null, this.data[pos.index]);
    }, this));
  };
  Storage.prototype.write = function(data, callback) {
    return process.nextTick(__bind(function() {
      return callback(null, new Position(this.data.push(data) - 1));
    }, this));
  };
  Storage.prototype.readRoot = function(callback) {
    return process.nextTick(__bind(function() {
      return callback(null, this.data[this.root_pos.index]);
    }, this));
  };
  Storage.prototype.writeRoot = function(root_pos, callback) {
    if (!isPosition(root_pos)) {
      return callback('pos should be a valid position');
    }
    return process.nextTick(__bind(function() {
      this.root_pos = root_pos;
      return callback(null);
    }, this));
  };
  Storage.prototype.inspect = function() {
    this.data.forEach(function(line, i) {
      return util.puts(i + ': ' + JSON.stringify(line));
    });
    return util.puts('Root : ' + JSON.stringify(this.root_pos));
  };
  Position = exports.Position = function(index) {
    this.index = index;
    return this;
  };
  /*
    Storage state
  */
  Storage.prototype.getState = function() {
    return {};
  };
  Storage.prototype.setState = function(state) {
    return true;
  };
  /*
    Compaction flow
  */
  Storage.prototype.beforeCompact = function(callback) {
    this._compactEdge = this.data.push('--------');
    return callback(null);
  };
  Storage.prototype.afterCompact = function(callback) {
    var i, _ref;
    for (i = 0, _ref = this._compactEdge; 0 <= _ref ? i < _ref : i > _ref; 0 <= _ref ? i++ : i--) {
      this.data[i] = 0;
    }
    return callback(null);
  };
}).call(this);
